-- Utils.lua: common lua used across multiple modules

local _,L = ...
local rematch = Rematch

-- takes an petid and returns what type of id it is ("pet" "species" or "leveling")
function rematch:GetIDType(id)
	if type(id)=="string" and id:match("^BattlePet%-%x%-%x%x%x%x%x%x%x%x%x%x%x%x$") then
		return "pet"
	elseif id==0 then
		return "leveling"
	elseif type(id)=="number" then
		return "species"
	end
end

--[[ Breed ]]

-- returns the addon that will provide breed data
function rematch:FindBreedSource()
	local addon
	for _,name in pairs({"BattlePetBreedID","PetTracker_Breeds","LibPetBreedInfo-1.0"}) do
		if not addon and IsAddOnLoaded(name) then
			addon = name
		end
	end
	if not addon then -- one of the sources is not loaded, try loading LibPetBreedInfo separately
		LoadAddOn("LibPetBreedInfo-1.0")
		if LibStub then
			for lib in LibStub:IterateLibraries() do
				if lib=="LibPetBreedInfo-1.0" then
					rematch.breedLib = LibStub("LibPetBreedInfo-1.0")
					addon = lib
					break
				end
			end
		end
	end
	rematch:GatherBreedNames() -- while here, gather names of the breeds indexed by breed filter
	return addon
end

-- returns the breed of the petID
function rematch:GetBreed(petID)
	local source = rematch.breedSource
	if source=="BattlePetBreedID" then
		return GetBreedID_Journal(petID) or ""
	elseif source=="PetTracker_Breeds" then
		local icon = PetTracker:GetBreedIcon((PetTracker.Journal:GetBreed(petID)),.95)
		if not icon or icon=="" then
			icon = "NEW"
		end
		return icon
	elseif source=="LibPetBreedInfo-1.0" then
		return rematch.breedLib:GetBreedName(rematch.breedLib:GetBreedByPetID(petID)) or "NEW"
	end
	return ""
end

-- called not only by the FindBreedSource function, but also if BPBID is used and BPBID_Options.format changes
function rematch:GatherBreedNames()
	wipe(rematch.breedNames)
	wipe(rematch.breedLookup)
	if rematch.breedSource=="BattlePetBreedID" then
		if BPBID_Options.format==1 then
			rematch.breedNames = {3,4,5,6,7,8,9,10,11,12}
		elseif BPBID_Options.format==2 then
			rematch.breedNames = {"3/13","4/14","5/15","6/16","7/17","8/18","9/19","10/20","11/21","12/22"}
		elseif BPBID_Options.format==3 then
			rematch.breedNames = {"B/B","P/P","S/S","H/H","H/P","P/S","H/S","P/B","S/B","H/B"}
		end
		-- since BPBID doesn't return a number, make a lookup to convert its names to a filter index
		for index,breedName in ipairs(rematch.breedNames) do
			rematch.breedLookup[breedName] = index
		end
		rematch.BPBIDFormat = BPBID_Options.format
	elseif rematch.breedSource=="PetTracker_Breeds" then
		-- PetTracker returns a number, so we can just use breed minus 2 to compare to filter settings
		for i=3,12 do
			tinsert(rematch.breedNames,format("%s %s",PetTracker:GetBreedIcon(i,.95),PetTracker:GetBreedName(i)))
		end
	end
end

-- returns the name of the team for display purposes
-- if color is true, prefixes white color code if it's a team with an npcID
function rematch:GetTeamTitle(key,color)
	local forNpc = type(key)=="number"
	local saved = RematchSaved
	if saved[key] and forNpc then
		local teamName = saved[key].teamName or format("NPC:%s",tostring(key))
		if color then
			return format("\124cffffffff%s\124r",teamName)
		else
			return teamName
		end
	else
		return key
	end
end

-- returns name of the given unit and npcID if it's an npc, or nil if unit doesn't exist
function rematch:GetUnitNameandID(unit)
	if UnitExists(unit) then
		local name = UnitName(unit)
		local npcID = tonumber((UnitGUID(unit) or ""):match(".-%-%d+%-%d+%-%d+%-%d+%-(%d+)"))
		if npcID and npcID~=0 then
			if rematch.notableRedirects[npcID] and not RematchSaved[npcID] then -- if this is a challenge post pet then return npcID of its post
				npcID = rematch.notableRedirects[npcID]
				return (rematch:GetNameFromNpcID(npcID)),npcID
			else
				return name,npcID -- this is an npc, return its name and npcID
			end
		else
			return name -- this is a player, return its name
		end
	end
end

function rematch:GetPetIcon(petID)
	local idType = rematch:GetIDType(petID)
	if idType=="pet" then
		return (select(9,C_PetJournal.GetPetInfoByPetID(petID)))
	elseif idType=="species" then
		return (select(2,C_PetJournal.GetPetInfoBySpeciesID(petID)))
	elseif idType=="leveling" then
		return rematch.levelingIcon
	else
		return "Interface\\Icons\\INV_Misc_QuestionMark"
	end
end

function rematch:GetPetName(petID)
	local idType = rematch:GetIDType(petID)
	if idType=="pet" then
		local _,customName,_,_,_,_,_,name = C_PetJournal.GetPetInfoByPetID(petID)
		return customName or name
	elseif idType=="species" then
		return (C_PetJournal.GetPetInfoBySpeciesID(petID))
	elseif idType=="leveling" then
		return L["Leveling Pet"]
	else
		return UNKNOWN
	end
end

function rematch:GetPetSpeciesID(petID)
	local idType = rematch:GetIDType(petID)
	local speciesID
	if idType=="pet" then
		return (C_PetJournal.GetPetInfoByPetID(petID))
	elseif idType=="species" then
		return petID
	end
end

-- DesensitizeText returns text in a literal (magic characters escaped) and case-insensitive format
local function literal(c) return "%"..c end
local function caseinsensitive(c) return format("[%s%s]",c:lower(),c:upper()) end
function rematch:DesensitizeText(text)
	if type(text)=="string" then
		return text:gsub("[%(%)%.%%%+%-%*%?%[%]%^%$]",literal):gsub("%a",caseinsensitive)
	end
end

-- use this instead of SetPetLoadOutInfo directly in case first slot changes (to resummon pet back)
function rematch:SlotPet(slot,petID)
	-- if KeepSummoned enable, note pet that's out and set up a timer to restore it
	if RematchSettings.KeepSummoned then
		rematch.preLoadCompanion = rematch.preLoadCompanion or C_PetJournal.GetSummonedPetGUID()
		rematch:RegisterEvent("UNIT_PET")
		rematch:StartTimer("RestoreTimeout",1,rematch.RestoreTimeout)
	end
	C_PetJournal.SetPetLoadOutInfo(slot,petID)
end

-- called from the RestoreTimeout timer, 
function rematch:RestoreCompanion()
	if not InCombatLockdown() then -- can't SummonPetByGUID in combat :(
		local nowSummoned = C_PetJournal.GetSummonedPetGUID()
		if not rematch.preLoadCompanion and nowSummoned then
			C_PetJournal.SummonPetByGUID(nowSummoned) -- something summoned, had nothing before
		elseif nowSummoned ~= rematch.preLoadCompanion then
			C_PetJournal.SummonPetByGUID(rematch.preLoadCompanion) -- something summoned different than before
		end
		rematch.preLoadCompanion = nil
	end
end
-- if no UNIT_PET fired, unregister it. apparently no pet was loaded
function rematch:RestoreTimeout()
	rematch.preLoadCompanion = nil
	rematch:UnregisterEvent("UNIT_PET")
end

-- hides all the pop-up and flyout widgets when doing any major event
-- can pass name of function "HideMenu" or "HideFlyout" to hide everything but that
-- if a menu item just clicked, it's automatically excluded from hiding (menu will take care of hiding itself)
local hideWidgetFuncs = {"HideMenu","HideFlyout","HidePetCard","HideTooltip"}
function rematch:HideWidgets(except)
	for _,funcName in pairs(hideWidgetFuncs) do
		if funcName~=except and rematch[funcName] and (funcName~="HideMenu" or not rematch:UIJustChanged()) then
			rematch[funcName]()
		end
	end
end

-- returns the pet on the cursor if there is one
function rematch:GetCursorPet()
	local hasPet,petID = GetCursorInfo()
	if hasPet=="battlepet" then
		return petID
	end
end

-- anchors frame to relativeTo depending on the quarter of the screen the reference frame
-- exists (will go up parents of relativeTo to find reference parent)
-- if no relativeTo given, it will choose RematchFrame or RematchJournal
-- if center is true it will anchor to reference frame's center
local references = { ["RematchFrame"]=true, ["UIParent"]=true }
function rematch:SmartAnchor(frame,relativeTo,center)
	if not relativeTo then
		relativeTo = RematchFrame -- will check if journal visible and change to it
	end

	local reference = relativeTo
	while not references[reference:GetName() or ""] do
		local parent = reference:GetParent()
		if not parent or parent==UIParent then
			break -- stop when we've run out of parents or reached UIParent
		else
			reference = parent
		end
	end
	if not reference then return end -- failed catastrophically

	-- pet card has a (usually) invisible titlebar; yspecial will nudge yoffset up 22 if anchoring to top
	local yspecial = frame==RematchPetCard and 22 or 0

	frame:ClearAllPoints()
	if center then -- simply center it to reference
		frame:SetPoint("CENTER",reference,"CENTER")
	else
		local corner = rematch:GetCorner(reference,UIParent)
		if corner=="BOTTOMLEFT" or reference==CollectionsJournal then
			frame:SetPoint("BOTTOMLEFT",relativeTo,"TOPRIGHT")
		elseif corner=="TOPLEFT" then
			frame:SetPoint("TOPLEFT",relativeTo,"BOTTOMRIGHT",0,yspecial)
		elseif corner=="TOPRIGHT" then
			frame:SetPoint("TOPRIGHT",relativeTo.Pet or relativeTo,"BOTTOMLEFT",0,yspecial)
		else 
			frame:SetPoint("BOTTOMRIGHT",relativeTo.Pet or relativeTo,"TOPLEFT")
		end
	end
end

-- To prevent tooltips and pet cards appearing when menus/pet cards/etc hide over an element,
-- any major event will mark the time it happened with timeUIChanged. Then this function
-- can be called to know whether an OnEnter should be run.
function rematch:UIJustChanged()
	return GetTime()==rematch.timeUIChanged
end

-- returns the corner of reference that frame is closest to
-- (if frame is closest to TOPRIGHT corner of reference, returns "TOPRIGHT")
-- used in SmartAnchor and anchoring RematchFrame to UIParent
-- if coords is true, the x,y offset from UIParent's BOTTOMLEFT is returned
function rematch:GetCorner(frame,reference,coords)
	local fx,fy = frame:GetCenter()
	local rx,ry = reference:GetCenter()
	local left, right, top, bottom
	if coords then
		left = frame:GetLeft()
		right = frame:GetRight()
		top = frame:GetTop()
		bottom = frame:GetBottom()
	end
	if not coords then
		ry=ry*1.2 -- raise y threshold up 20% of reference if not returning coords (to favor anchoring upwards)
	end
	if fx<rx and fy<ry then -- bottomleft
		return "BOTTOMLEFT",left,bottom
	elseif fx<rx and fy>ry then -- topleft
		return "TOPLEFT",left,top
	elseif fx>rx and fy>ry then -- topright
		return "TOPRIGHT",right,top
	else -- bottomright (or dead center)
		return "BOTTOMRIGHT",right,bottom
	end
end

-- adds a leveling border to a button (only 3 main loadouts and queue leveling slot uses this)
function rematch:AddLevelingBorder(button)
	button.Leveling = button:CreateTexture(nil,"BACKGROUND")
	local cx,cy = button:GetSize()
	button.Leveling:SetSize(cx+10,cy+10)
	button.Leveling:SetPoint("CENTER")
	button.Leveling:SetTexture("Interface\\PetBattles\\PetBattle-GoldSpeedFrame")
	button.Leveling:SetTexCoord(0.1171875,0.7421875,0.1171875,0.734375)
end

-- adds frameName to UISpecialFrames if value is true, removes it if value is false
function rematch:SetESCable(frameName,value)
	if value and not tContains(UISpecialFrames,frameName) then
		tinsert(UISpecialFrames,frameName)
	elseif not value then
		for i=#UISpecialFrames,1,-1 do
			if UISpecialFrames[i]==frameName then
				tremove(UISpecialFrames,i)
			end
		end
	end
end

function rematch:print(...)
	print(format("%s%s:\124r",rematch.hexGold,L["Rematch"]),...)
end

-- these two functions find the highest framelevel of a frame and its children
-- highestLevel = rematch:FindHighestFrameLevel(startingFrame)
-- RematchJournal is excluded
local function findHigherFrameLevel(...)
	local numFrames = select("#",...)
	for i=1,numFrames do
		local frame = select(i,...)
		if not RematchJournal or frame~=RematchJournal then
			local level = frame:GetFrameLevel()
			if level<500 then -- if something badly wants to be higher framelevel (like PBT) let it
				rematch.highestFrameLevel = max(rematch.highestFrameLevel,level)
			end
			findHigherFrameLevel(frame:GetChildren())
		end
	end
end
-- PBT's TeamFrame.lua creates a teamMovementFrame with 10k framelevel lol
function rematch:FindHighestFrameLevel(frame)
	rematch.highestFrameLevel = 0
	findHigherFrameLevel(frame)
	return rematch.highestFrameLevel
end

--[[ List scrolling ]]

function rematch:ListScrollToTop(scrollFrame)
	scrollFrame.scrollBar:SetValue(0)
	PlaySound("UChatScrollButton")
end

function rematch:ListScrollToBottom(scrollFrame)
	scrollFrame.scrollBar:SetValue(scrollFrame.range)
	PlaySound("UChatScrollButton")
end

function rematch:ListScrollToIndex(scrollFrame,index)
	if index then
		if scrollFrame.scrollBar:IsEnabled() then
			local buttons = scrollFrame.buttons
			local height = math.max(0,floor(scrollFrame.buttonHeight*(index-((scrollFrame:GetHeight()/scrollFrame.buttonHeight))/2)))
			HybridScrollFrame_SetOffset(scrollFrame,height)
			scrollFrame.scrollBar:SetValue(height)
		else
			rematch:ListScrollToTop(scrollFrame)
		end
	end
end

-- does a flashing glow in a listbutton (after 0.05 seconds to let it update)
function rematch:ListBling(scrollFrame,var,value)
	rematch.blingScrollFrame = scrollFrame
	rematch.blingVar = var
	rematch.blingValue = value
	C_Timer.After(0.05,rematch.ListDoBling)
end

-- does the actual glow effect after 0.05 seconds
function rematch:ListDoBling()
	local scrollFrame = rematch.blingScrollFrame
	if scrollFrame then
		for _,button in ipairs(scrollFrame.buttons) do
			if button[rematch.blingVar]==rematch.blingValue then
				local bling = RematchPetListBling
				bling:SetParent(button)
				bling:SetFrameLevel(button:GetFrameLevel()+5)
				bling:SetAllPoints(true)
				bling:Show()
				return
			end
		end
	end
	rematch.blingScrollFrame = nil
	rematch.blingVar = nil
	rematch.blingValue = nil
end

-- takes a petType (1-10) and returns a text string for the icon (20x20) with the thin circle border
function rematch:PetTypeAsText(petType,size)
	size = size or 16
	local suffix = PET_TYPE_SUFFIX[petType]
	return suffix and format("\124TInterface\\PetBattles\\PetIcon-%s:%d:%d:0:0:128:256:102:63:129:168\124t",suffix,size,size) or "?"
end

function rematch:SetTopToggleButton(button,up)
	button.up = up
	rematch.TopToggleButtonOnMouseUp(button)
	if up then
--		button.Icon:SetTexture("Interface\\Buttons\\UI-MicroStream-Yellow")
--		button.Icon:SetTexCoord(0,1,1,0)
--		button.Icon:SetTexCoord(1,0,0,0,1,1,0,1) -- 90 degrees (up)
	else
--		button.Icon:SetTexture("Interface\\Buttons\\UI-MicroStream-Yellow")
--		button.Icon:SetTexCoord(0,1,0,1)
--		button.Icon:SetTexCoord(0,1,1,1,0,0,1,0) -- -90 degrees (down)
	end
end

function rematch:TopToggleButtonOnMouseDown(button)
	if self:IsEnabled() then
		self.TopLeft:SetTexture("Interface\\Buttons\\UI-Silver-Button-Down")
		self.TopRight:SetTexture("Interface\\Buttons\\UI-Silver-Button-Down")
		self.BottomLeft:SetTexture("Interface\\Buttons\\UI-Silver-Button-Down")
		self.BottomRight:SetTexture("Interface\\Buttons\\UI-Silver-Button-Down")
		self.MiddleLeft:SetTexture("Interface\\Buttons\\UI-Silver-Button-Down")
		self.MiddleRight:SetTexture("Interface\\Buttons\\UI-Silver-Button-Down")
		if self.up then
			self.Icon:SetPoint("CENTER",-1,0)
		else
			self.Icon:SetPoint("CENTER",-1,-2)
		end
	end
end

function rematch:TopToggleButtonOnMouseUp(button)
	self.TopLeft:SetTexture("Interface\\Buttons\\UI-Silver-Button-Up")
	self.TopRight:SetTexture("Interface\\Buttons\\UI-Silver-Button-Up")
	self.BottomLeft:SetTexture("Interface\\Buttons\\UI-Silver-Button-Up")
	self.BottomRight:SetTexture("Interface\\Buttons\\UI-Silver-Button-Up")
	self.MiddleLeft:SetTexture("Interface\\Buttons\\UI-Silver-Button-Up")
	self.MiddleRight:SetTexture("Interface\\Buttons\\UI-Silver-Button-Up")
	if self.up then
		self.Icon:SetPoint("CENTER",0,1)
		self.Icon:SetTexCoord(1,0,0,0,1,1,0,1) -- 90 degrees (up)
	else
		self.Icon:SetPoint("CENTER",0,-1)
		self.Icon:SetTexCoord(0,1,1,1,0,0,1,0) -- -90 degrees (down)
	end
end

-- adjusts the scale of a frame depending on settings.SmallerWindow and if
-- the regular frame is up; if force is true then adjusts regardless if frame
-- is visible or not.
function rematch:AdjustScale(frame,force)
	if RematchSettings.SmallerWindow and (rematch.Frame:IsVisible() or force) then
		frame:SetScale(0.8)
	else
		frame:SetScale(1)
	end
end

-- to be called when some action (prompt to load->with window; auto load->show etc)
-- wants to show the window. if neither standalone or journal are up, it shows
-- the the PreferredMode (1=minimized, 2=maximized, 3=journal)
function rematch:AutoShow()
	local frame = rematch.Frame
	local journal = rematch.Journal
	local settings = RematchSettings
	local mode = settings.PreferredMode
	if not frame:IsVisible() and not journal:IsVisible() then
		if mode==1 then
			settings.Minimized = true
			rematch.Frame.Toggle()
		elseif mode==2 then
			settings.Minimized = nil
			rematch.Frame.Toggle()
		elseif mode==3 then
			settings.UseDefaultJournal = nil
			ToggleCollectionsJournal(2)
		end
	end
end

-- for minimap and tab bindings, instead of frame:Toggle() to toggle the standalone
-- window, this toggles the preferred mode.
function rematch:AutoToggle()

end